#include "dispatcher.h"

Dispatcher::Dispatcher(Uploader *uploader, QObject *parent) :
    QObject(parent), m_client(NULL), m_uploader(uploader)
{
}

rfbBool Dispatcher::MallocFrameBuffer(rfbClient *client)
{
    Dispatcher *dispatcher = (Dispatcher *) rfbClientGetClientData(client, 0);
    emit dispatcher->resizeFramebuffer(QSize(client->width, client->height));
    return TRUE;
}

void Dispatcher::GotCopyRect(rfbClient *client, const int src_x, const int src_y, const int w, const int h, const int dest_x, const int dest_y)
{
    Dispatcher *dispatcher = (Dispatcher *) rfbClientGetClientData(client, 0);
    dispatcher->uploader()->gotCopy(src_x, src_y, w, h, dest_x, dest_y);
}

void Dispatcher::GotFillRect(rfbClient *client, int x, int y, int w, int h, uint32_t colour)
{
    Dispatcher *dispatcher = (Dispatcher *) rfbClientGetClientData(client, 0);
    int red = (colour >> client->format.redShift) & client->format.redMax;
    int green = (colour >> client->format.greenShift) & client->format.greenMax;
    int blue = (colour >> client->format.blueShift) & client->format.blueMax;
    dispatcher->uploader()->gotFill(x, y, w, h, red, green, blue);
}

void Dispatcher::GotBitmap(rfbClient *client, const uint8_t *buffer, int x, int y, int w, int h)
{
    Dispatcher *dispatcher = (Dispatcher *) rfbClientGetClientData(client, 0);
    // We have to make a deep copy because libvnc might free the buffer as soon as this function returns
    QImage image(buffer, w, h, QImage::Format_RGBX8888);
    dispatcher->uploader()->gotBitmap(image.convertToFormat(QImage::Format_RGB888), x, y, w, h);
}

rfbBool Dispatcher::GotJpeg(rfbClient *client, const uint8_t *buffer, int length, int x, int y, int w, int h)
{
    Dispatcher *dispatcher = static_cast<Dispatcher *>(rfbClientGetClientData(client, 0));
    dispatcher->uploader()->gotJpeg();
    Jpeg *jpeg = new Jpeg({(unsigned char*)buffer, length, x, y, w, h});
    dispatcher->queue()->enqueue(jpeg);
    return TRUE;
}

void Dispatcher::FinishedFrameBufferUpdate(rfbClient *client)
{
    Dispatcher *dispatcher = (Dispatcher *) rfbClientGetClientData(client, 0);
    dispatcher->uploader()->finishedUpdate();
}

bool Dispatcher::mouseEvent(const int x, const int y, const bool left, const bool middle, const bool right, const bool up, const bool down)
{
    if(m_client == NULL)
        return false;

    int buttonMask = 0;
    if(left) buttonMask |= rfbButton1Mask;
    if(right) buttonMask |= rfbButton2Mask;
    if(middle) buttonMask |= rfbButton3Mask;
    if(up) buttonMask |= rfbButton4Mask;
    if(down) buttonMask |= rfbButton5Mask;
    SendPointerEvent(m_client, x, y, buttonMask);
    return true;
}

int Dispatcher::bitsPerPixel()
{
    if(m_client == NULL)
        return 0;
    return m_client->format.bitsPerPixel;
}

ConcurrentQueue<Jpeg *, 256u> *Dispatcher::queue()
{
    return &m_queue;
}

Uploader *Dispatcher::uploader() const
{
    return m_uploader;
}

void Dispatcher::open(const QString host, const int port, const int width, const int height)
{
    QByteArray tmp = host.toLocal8Bit();
    m_client = rfbGetClient(8, 3, 4);

    m_client->canHandleNewFBSize = FALSE;
    m_client->serverHost = new char[tmp.size() + 1];
    strcpy(m_client->serverHost, tmp.constData());
    m_client->serverPort = port;
    m_client->MallocFrameBuffer = Dispatcher::MallocFrameBuffer;
    m_client->GotCopyRect = Dispatcher::GotCopyRect;
    m_client->GotFillRect = Dispatcher::GotFillRect;
    m_client->GotBitmap = Dispatcher::GotBitmap;
    m_client->GotJpeg = Dispatcher::GotJpeg;
    m_client->FinishedFrameBufferUpdate = Dispatcher::FinishedFrameBufferUpdate;
    m_client->appData.useRemoteCursor = TRUE;

    // Set the format requested by the user
    m_client->width = width;
    m_client->height = height;
    //m_client->appData.compressLevel = 0;
    //m_client->appData.qualityLevel = 0;
#ifdef __BIG_ENDIAN__
    m_client->format.bigEndian = TRUE;
#else
    m_client->format.bigEndian = FALSE;
#endif
    // The Raspberry Pi supports 8 bit paletted images and RGB422 too but only through extensions
    m_client->format.trueColour = TRUE;
    m_client->format.bitsPerPixel = 32;

    rfbClientSetClientData(m_client, 0, this);

    if(!rfbInitClient(m_client, NULL, NULL)) {
        emit error();
        return;
    }

    // Stick to the format sent by the server if we can render it easily, set a new one otherwise
    //
    // Formats OpenGL ES 2.0 can (should be able to) handle:
    // Internal Format     External Format Type                    Bytes per Pixel
    // ---------------     --------------- ----                    ---------------
    // RGBA                RGBA             UNSIGNED_BYTE          4
    // RGB                 RGB              UNSIGNED_BYTE          3
    // RGBA                RGBA             UNSIGNED_SHORT_4_4_4_4 2
    // RGBA                RGBA             UNSIGNED_SHORT_5_5_5_1 2
    // RGB                 RGB              UNSIGNED_SHORT_5_6_5   2
    // The Raspberry Pi supports BGRA too but only through extensions
    //
    // JPEG is RGB UNSIGNED_BYTE
    // Supported raw formats:
    // 32 bit: RGB888
    // 16 bit: RGB444, RGB555, RGB565

    int message = WaitForMessage(m_client, 1000);
    if(message < 0) {
        terminate();
        emit error();
        return;
    }
    refresh();
}
void Dispatcher::refresh()
{
    if(m_client == NULL)
        return;

    int message = 0;

    // Wait until we get something
    while(message == 0) {
        message = WaitForMessage(m_client, 1000);
        if(message < 0) {
            terminate();
            emit error();
            return;
        }
    }

    if(!HandleRFBServerMessage(m_client)) {
        terminate();
        emit error();
        return;
    }
}

void Dispatcher::terminate()
{
    if(m_client == NULL)
        return;

    rfbClientCleanup(m_client);
}
