#include "qvnc.h"

QVnc::QVnc(QQuickItem *parent) :
    QQuickFramebufferObject(parent), m_decoders(DECODER_COUNT), m_threads(DECODER_COUNT + 1)
{
    setAcceptedMouseButtons(Qt::AllButtons);
    setAcceptHoverEvents(true);
    setTextureFollowsItemSize(false);

    for(int i = 0; i < m_threads.length(); i++)
        m_threads[i] = new QThread(this);

    connect(this, &QVnc::windowChanged, this, &QVnc::winChanged);

    m_uploader = new Uploader;
    m_uploader->moveToThread(m_threads[0]);

    connect(m_threads[0], &QThread::finished, m_uploader, &QObject::deleteLater);
    connect(m_uploader, &Uploader::uploadFinished, this, &QVnc::update);

    m_dispatcher = new Dispatcher(m_uploader);
    m_dispatcher->moveToThread(m_threads[0]);

    connect(m_threads[0], &QThread::finished, m_dispatcher, &QObject::deleteLater);
    connect(this, &QVnc::open_connection, m_dispatcher, &Dispatcher::open);
    connect(this, &QVnc::terminate_connection, m_dispatcher, &Dispatcher::terminate);
    connect(m_dispatcher, &Dispatcher::error, this, &QVnc::dispatch_error);
    connect(m_uploader, &Uploader::uploadFinished, m_dispatcher, &Dispatcher::refresh);

    int i = 1;
    foreach(JpegDecoder* d, m_decoders) {
        d = new JpegDecoder(m_dispatcher->queue());
        d->moveToThread(m_threads[i]);
        connect(m_threads[i], &QThread::finished, d, &QObject::deleteLater);
        connect(d, &JpegDecoder::finished, m_uploader, &Uploader::finishedJpeg);
        connect(this, &QVnc::operate, d, &JpegDecoder::operate);
        i++;
    }

    for(int i = 0; i < m_threads.length(); i++)
        m_threads.at(i)->start();

    emit operate();
}

QVnc::~QVnc()
{
    foreach(QThread* t, m_threads) {
        t->quit();
        t->wait();
    }
}

void QVnc::open()
{
    emit open_connection(this->m_host, this->m_port, (int) this->width(), (int) this->height());
}

void QVnc::terminate()
{
    emit terminate_connection();
}

void QVnc::setHost(QString host)
{
    if (m_host == host)
        return;

    m_host = host;
    emit hostChanged(host);
}

void QVnc::setPort(int port)
{
    if (m_port == port)
        return;

    m_port = port;
    emit portChanged(port);
}

void QVnc::dispatch_error()
{
    qDebug() << "error";
}

void QVnc::winChanged(QQuickWindow *window)
{
    if(window != NULL) {
        connect(window, &QQuickWindow::openglContextCreated, m_uploader, &Uploader::createContext, Qt::BlockingQueuedConnection);
        connect(window, &QQuickWindow::frameSwapped, m_uploader, &Uploader::frameSwapped);
    }
}

QQuickFramebufferObject::Renderer *QVnc::createRenderer() const
{
    VncRenderer *renderer = new VncRenderer(window(),
                                            m_uploader->vertices(),
                                            m_uploader->indices(),
                                            m_uploader->texture(),
                                            m_uploader->drawCount());
    connect(renderer, &VncRenderer::FBOTextureChanged, m_uploader, &Uploader::changeFBOTexture);
    return renderer;
}

QString QVnc::host() const
{
    return m_host;
}

int QVnc::port() const
{
    return m_port;
}

void QVnc::mousePressEvent(QMouseEvent *event)
{
    Qt::MouseButtons buttons = event->buttons();
    m_lastMouseEvent = event->timestamp();
    if(m_dispatcher->mouseEvent(event->x(),
                                event->y(),
                                buttons & Qt::LeftButton,
                                buttons & Qt::MiddleButton,
                                buttons & Qt::RightButton,
                                false,
                                false))
        event->accept();
    else
        event->ignore();
}
void QVnc::mouseReleaseEvent(QMouseEvent *event)
{
    Qt::MouseButtons buttons = event->buttons();
    m_lastMouseEvent = event->timestamp();
    if(m_dispatcher->mouseEvent(event->x(),
                                event->y(),
                                buttons & Qt::LeftButton,
                                buttons & Qt::MiddleButton,
                                buttons & Qt::RightButton,
                                false,
                                false))
        event->accept();
    else
        event->ignore();
}

void QVnc::wheelEvent(QWheelEvent *event)
{
    Qt::MouseButtons buttons = event->buttons();
    QPoint rotated = event->angleDelta();
    m_lastMouseEvent = event->timestamp();
    if(m_dispatcher->mouseEvent(event->x(),
                                event->y(),
                                buttons & Qt::LeftButton,
                                buttons & Qt::MiddleButton,
                                buttons & Qt::RightButton,
                                rotated.x() > 0,
                                rotated.x() < 0))
        event->accept();
    else
        event->ignore();
}

void QVnc::mouseMoveEvent(QMouseEvent *event)
{
    // Filter mouse move events, we don't need 100+ of them in a second
    if((event->timestamp() - m_lastMouseEvent) > 33) {
        Qt::MouseButtons buttons = event->buttons();
        m_lastMouseEvent = event->timestamp();
        if(m_dispatcher->mouseEvent(event->x(),
                                    event->y(),
                                    buttons & Qt::LeftButton,
                                    buttons & Qt::MiddleButton,
                                    buttons & Qt::RightButton,
                                    false,
                                    false))
            event->accept();
        else
            event->ignore();
    }
}

void QVnc::hoverMoveEvent(QHoverEvent *event)
{
    // Qt bug: timestamp is zero for QHoverEvent
    // if((event->timestamp() - m_lastMouseEvent) > 33) {
    //    m_lastMouseEvent = event->timestamp();
        QPoint position = event->pos();
        if(m_dispatcher->mouseEvent(position.x(),
                                    position.y(),
                                    false,
                                    false,
                                    false,
                                    false,
                                    false))
            event->accept();
        else
            event->ignore();
    // }
}

void QVnc::keyPressEvent(QKeyEvent *event)
{
}

void QVnc::keyReleaseEvent(QKeyEvent *event)
{
}
