#include "uploader.h"
#include <cstring>

Uploader::Uploader(QObject *parent) :
    QObject(parent), m_context(this), m_indices(QOpenGLBuffer::IndexBuffer),
    m_uploadVertices(new QOpenGLBuffer*), m_renderVertices(new QOpenGLBuffer*),
    m_uploadTexture(new GLuint*), m_renderTexture(new GLuint*), m_program(this), m_jpegs(0),
    m_atlasX(0), m_atlasY(0), m_atlasRowHeight(0), m_atlasLastWidth(0), m_copyIndex(0)
{
    // "applications must ensure that create() is only called on the main (GUI) thread"
    m_surface.create();
    m_drawCount = new unsigned int(0);
}

Uploader::~Uploader()
{
    m_program.release();
}

void Uploader::createContext(QOpenGLContext *context)
{
    m_context.setShareContext(context);
    m_context.create();
    m_context.makeCurrent(&m_surface);

    m_functions = m_context.functions();
    m_functions->glEnable(GL_CULL_FACE);

    m_program.addShaderFromSourceFile(QOpenGLShader::Vertex, "copy_shader.vsh");
    m_program.addShaderFromSourceFile(QOpenGLShader::Fragment, "copy_shader.fsh");
    m_program.bind();
    m_program.setUniformValue("texture", 0);
    QMatrix4x4 ortho;
    ortho.ortho(0, TEXTURE_WIDTH, 0, TEXTURE_HEIGHT, -1, 1);
    // m_program.setUniformValue("ortho", ortho);
    m_program.enableAttributeArray("position");
    m_program.enableAttributeArray("texcoord");
    m_functions->glVertexAttribPointer(m_program.attributeLocation("position"),
                                       2,
                                       GL_UNSIGNED_SHORT,
                                       GL_FALSE,
                                       sizeof(Copy),
                                       (void*) 0);
    m_functions->glVertexAttribPointer(m_program.attributeLocation("texcoord"),
                                       2,
                                       GL_UNSIGNED_SHORT,
                                       GL_FALSE,
                                       sizeof(Copy),
                                       (void*) (2 * sizeof(GLushort)));

    m_indices.create();
    m_indices.bind();
    m_indices.allocate(ind.i, (VERTEX_BUFFER * 6 - 2) * sizeof(unsigned short));

    m_copyBuffer.create();
    m_copyBuffer.bind();
    m_copyBuffer.allocate(COPY_BUFFER * sizeof(Copy));
    m_copyPointer = (Copy*) m_copyBuffer.map(QOpenGLBuffer::WriteOnly);

    *m_renderVertices = new QOpenGLBuffer;
    (*m_renderVertices)->setUsagePattern(QOpenGLBuffer::DynamicDraw);
    (*m_renderVertices)->create();
    (*m_renderVertices)->bind();
    (*m_renderVertices)->allocate(VERTEX_BUFFER * sizeof(Vertex));

    *m_uploadVertices = new QOpenGLBuffer;
    (*m_uploadVertices)->setUsagePattern(QOpenGLBuffer::DynamicDraw);
    (*m_uploadVertices)->create();
    (*m_uploadVertices)->bind();
    (*m_uploadVertices)->allocate(VERTEX_BUFFER * sizeof(Vertex));

    m_vertexPointer = (Vertex*) (*m_uploadVertices)->map(QOpenGLBuffer::WriteOnly);
    m_vertexIndex = 0;

    m_functions->glEnable(GL_TEXTURE_2D);
    m_functions->glActiveTexture(GL_TEXTURE0);
    m_functions->glGenTextures(2, m_texId);
    for(size_t i = 0; i < 2; i++) {
        m_functions->glBindTexture(GL_TEXTURE_2D, m_texId[i]);
        m_functions->glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, TEXTURE_WIDTH, TEXTURE_HEIGHT, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
    }

    *m_uploadTexture = m_texId;
    *m_renderTexture = m_texId + 1;

    m_functions->glBindTexture(GL_TEXTURE_2D, **m_uploadTexture);

    m_functions->glGenFramebuffers(1, &m_FBOId);
    m_functions->glBindFramebuffer(GL_FRAMEBUFFER, m_FBOId);
    m_functions->glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, **m_uploadTexture, 0);
    m_functions->glViewport(0, 0, TEXTURE_WIDTH, TEXTURE_HEIGHT);

    m_functions->glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
}

void Uploader::gotFill(const GLushort x, const GLushort y,
                       const GLushort width, const GLushort height,
                       const GLubyte red, const GLubyte green, const GLubyte blue)
{
    if(m_vertexIndex + 4 < VERTEX_BUFFER) {
        // bottom left
        m_vertexPointer[m_vertexIndex++] = Vertex({x, (GLushort) (y + height), red, green, blue, 0xFF, 0, 0});
        // top left
        m_vertexPointer[m_vertexIndex++] = Vertex({x, y, red, green, blue, 0xFF, 0, 0});
        // bottom right
        m_vertexPointer[m_vertexIndex++] = Vertex({(GLushort) (x + width), (GLushort) (y + height), red, green, blue, 0xFF, 0, 0});
        // top right
        m_vertexPointer[m_vertexIndex++] = Vertex({(GLushort) (x + width), y, red, green, blue, 0xFF, 0, 0});
    } else {
        qDebug() << "Out of memory";
    }
}

void Uploader::gotCopy(const GLushort src_x, const GLushort src_y,
                       const GLushort width, const GLushort height,
                       const GLushort dest_x, const GLushort dest_y)
{
    if(!refreshAtlas(width, height))
        return;

    // copy from source
    if(m_copyIndex + 4 < COPY_BUFFER) {
        m_copyPointer[m_copyIndex++] = Copy({m_atlasX, m_atlasY, src_x, (GLushort) (src_y + height)});
        m_copyPointer[m_copyIndex++] = Copy({m_atlasX, (GLushort) (m_atlasY + height), src_x, src_y});
        m_copyPointer[m_copyIndex++] = Copy({(GLushort) (m_atlasX + width), m_atlasY, (GLushort) (src_x + width), (GLushort) (src_y + height)});
        m_copyPointer[m_copyIndex++] = Copy({(GLushort) (m_atlasX + width), (GLushort) (m_atlasY + height), (GLushort) (src_x + width), src_y});
    } else {
        qDebug() << "Out of memory";
    }

    // copy to destination
    if(m_vertexIndex + 4 < VERTEX_BUFFER) {
        // bottom left
        m_vertexPointer[m_vertexIndex++] = Vertex({dest_x, (GLushort) (dest_y + height), 0, 0, 0, 0, m_atlasX, m_atlasY});
        // top left
        m_vertexPointer[m_vertexIndex++] = Vertex({dest_x, dest_y, 0, 0, 0, 0, m_atlasX, (GLushort) (m_atlasY + height)});
        // bottom right
        m_vertexPointer[m_vertexIndex++] = Vertex({(GLushort) (dest_x + width), (GLushort) (dest_y + height), 0, 0, 0, 0, (GLushort) (m_atlasX + width), m_atlasY});
        // top right
        m_vertexPointer[m_vertexIndex++] = Vertex({(GLushort) (dest_x + width), dest_y, 0, 0, 0, 0, (GLushort) (m_atlasX + width), (GLushort) (m_atlasY + height)});
    } else {
        qDebug() << "Out of memory";
    }
}

void Uploader::gotBitmap(const QImage &image,
                         const GLushort x, const GLushort y,
                         const GLushort width, const GLushort height)
{
    if(!refreshAtlas(width, height))
        return;

    m_functions->glTexSubImage2D(GL_TEXTURE_2D, 0, m_atlasX, m_atlasY, width, height, GL_RGB, GL_UNSIGNED_BYTE, image.constBits());

    if(m_vertexIndex + 4 < VERTEX_BUFFER) {
        // bottom left
        m_vertexPointer[m_vertexIndex++] = Vertex({x, (GLushort) (y + height), 0, 0, 0, 0, m_atlasX, (GLushort) (m_atlasY + height)});
        // top left
        m_vertexPointer[m_vertexIndex++] = Vertex({x, y, 0, 0, 0, 0, m_atlasX, m_atlasY});
        // bottom right
        m_vertexPointer[m_vertexIndex++] = Vertex({(GLushort) (x + width), (GLushort) (y + height), 0, 0, 0, 0, (GLushort) (m_atlasX + width), (GLushort) (m_atlasY + height)});
        // top right
        m_vertexPointer[m_vertexIndex++] = Vertex({(GLushort) (x + width), y, 0, 0, 0, 0, (GLushort) (m_atlasX + width), m_atlasY});
    } else {
        qDebug() << "Out of memory";
    }
}

void Uploader::gotJpeg()
{
    m_jpegs++;
}

void Uploader::finishedJpeg(const unsigned char *image,
                            const GLushort x, const GLushort y,
                            const GLushort width, const GLushort height)
{
    if(!refreshAtlas(width, height))
        return;

    m_jpegs--;

    m_functions->glTexSubImage2D(GL_TEXTURE_2D, 0, m_atlasX, m_atlasY, width, height, GL_RGB, GL_UNSIGNED_BYTE, image);
    delete[] image;

    if(m_vertexIndex + 4 < VERTEX_BUFFER) {
        // bottom left
        m_vertexPointer[m_vertexIndex++] = Vertex({x, (GLushort) (y + height), 0, 0, 0, 0, m_atlasX, (GLushort) (m_atlasY + height)});
        // top left
        m_vertexPointer[m_vertexIndex++] = Vertex({x, y, 0, 0, 0, 0, m_atlasX, m_atlasY});
        // bottom right
        m_vertexPointer[m_vertexIndex++] = Vertex({(GLushort) (x + width), (GLushort) (y + height), 0, 0, 0, 0, (GLushort) (m_atlasX + width), (GLushort) (m_atlasY + height)});
        // top right
        m_vertexPointer[m_vertexIndex++] = Vertex({(GLushort) (x + width), y, 0, 0, 0, 0, (GLushort) (m_atlasX + width), m_atlasY});
    } else {
        qDebug() << "Out of memory";
    }

    if(m_jpegs == 0 && m_finished && m_swapped) {
        startRendering();
    }
}

// TODO: the rendering thread might swap buffers before the startRendering signal reaches it
void Uploader::frameSwapped()
{
    m_swapped = true;
    if(m_finished) {
        copyRectangles();
        if(m_jpegs == 0) {
            startRendering();
        }
    }
}

void Uploader::finishedUpdate()
{
    m_finished = true;
    if(m_swapped) {
        copyRectangles();
        if(m_jpegs == 0)
            startRendering();
    }
}

bool Uploader::refreshAtlas(const int width, const int height)
{
    if(m_atlasX + m_atlasLastWidth + width <= TEXTURE_WIDTH) {
        if(m_atlasY + height > TEXTURE_HEIGHT) {
            qDebug() << "Out of memory";
            return false;
        }

        m_atlasX += m_atlasLastWidth;
        m_atlasLastWidth = width;

        if(height > m_atlasRowHeight)
            m_atlasRowHeight = height;
    } else {
        if(m_atlasY + m_atlasRowHeight + height > TEXTURE_HEIGHT) {
            qDebug() << "Out of memory";
            return false;
        }

        m_atlasX = 0;
        m_atlasY += m_atlasRowHeight;
        m_atlasRowHeight = height;
        m_atlasLastWidth = width;
    }
    return true;
}

void Uploader::startRendering()
{
    m_functions->glFinish();
    (*m_uploadVertices)->unmap();
    std::swap(*m_uploadVertices, *m_renderVertices);
    (*m_uploadVertices)->bind();
    m_vertexPointer = (Vertex*) (*m_uploadVertices)->map(QOpenGLBuffer::WriteOnly);
    if(m_vertexIndex > 0)
        *m_drawCount = (m_vertexIndex / 4) * 6 - 2;
    else
        *m_drawCount = 0;
    m_vertexIndex = 0;

    std::swap(*m_uploadTexture, *m_renderTexture);

    m_atlasX = 0;
    m_atlasY = 0;
    m_atlasRowHeight = 0;
    m_atlasLastWidth = 0;

    m_swapped = false;
    m_finished = false;
    m_functions->glBindTexture(GL_TEXTURE_2D, **m_uploadTexture);
    m_functions->glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, **m_uploadTexture, 0);
    emit uploadFinished();
}

void Uploader::copyRectangles()
{
    if(m_copyIndex > 0) {
        // m_functions->glActiveTexture(GL_TEXTURE1);
        m_functions->glBindTexture(GL_TEXTURE_2D, m_FBOTexture);
        m_functions->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        m_functions->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        m_functions->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        m_functions->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        m_copyBuffer.bind();
        m_copyBuffer.unmap();

        m_indices.bind();

        m_functions->glDrawElements(GL_TRIANGLE_STRIP, (m_copyIndex / 4) * 6 - 2, GL_UNSIGNED_SHORT, (GLvoid*) 0);
        m_functions->glFinish();

        m_copyPointer = (Copy*) m_copyBuffer.map(QOpenGLBuffer::WriteOnly);

        (*m_uploadVertices)->bind();
        m_functions->glBindTexture(GL_TEXTURE_2D, **m_uploadTexture);
        // m_functions->glActiveTexture(GL_TEXTURE0);
        m_copyIndex = 0;
    }
}

QOpenGLBuffer **Uploader::vertices() const
{
  return m_renderVertices;
}

GLuint **Uploader::texture() const
{
    return m_renderTexture;
}

QOpenGLBuffer Uploader::indices() const
{
    return m_indices;
}

unsigned int *Uploader::drawCount() const
{
    return m_drawCount;
}

void Uploader::changeFBOTexture(const unsigned int FBOTexture, const QMatrix4x4 &ortho)
{
    /*m_functions->glActiveTexture(GL_TEXTURE1);
    m_functions->glBindTexture(GL_TEXTURE_2D, m_FBOTexture);
    m_functions->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    m_functions->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    m_functions->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    m_functions->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    m_functions->glActiveTexture(GL_TEXTURE0);*/

    m_FBOTexture = FBOTexture;
    m_program.setUniformValue("texortho", ortho);
}
