/*
 * Design decisions:
 * - There is no way to access texture memory directly, copying is required.
 *   See: https://www.raspberrypi.org/forums/viewtopic.php?f=67&t=25959
 * - libvnc hooks free the memory buffer after they return (except for gotJpeg)
 *   Solutions:
 *    1) copy images to separate textures
 *       Pro: simple
 *       Con: multiple textures (slow)
 *    2) make deep copies of the images and move them to a single texture when decoding finished
 *       Pro: optimal texture atlas
 *       Con: deep copy
 *    3) first-fit allocation
 */

#ifndef UPLOADER_H
#define UPLOADER_H

#define TEXTURE_WIDTH 2048
#define TEXTURE_HEIGHT 2048
#define COPY_BUFFER 1024
#define VERTEX_BUFFER 1024

#include <utility>

#include <QObject>
#include <QQuickWindow>
#include <QDebug>
#include <QVector>
#include <QOffscreenSurface>
#include <QOpenGLContext>
#include <QOpenGLFunctions>
#include <QOpenGLBuffer>
#include <QOpenGLShaderProgram>
#include <QOpenGLFramebufferObject>
#include <QSharedPointer>
#include <QtMath>

struct Indices
{
    // 4 corner/rectangle and the first and last one duplicated except for the first and last rectangle
    unsigned short i[(VERTEX_BUFFER / 4) * 6 - 2];
    constexpr Indices() : i()
    {
        i[0] = 0;
        auto index = 1;
        for(auto value = 1; value < VERTEX_BUFFER; value++) {
            i[index++] = value;
            if(value % 4 == 0 || value % 4 == 3)
                i[index++] = value;
        }
    }
};
constexpr Indices ind;

struct Vertex
{
    GLushort x;
    GLushort y;
    GLubyte red;
    GLubyte green;
    GLubyte blue;
    GLubyte alpha;
    GLushort texX;
    GLushort texY;
};

struct Copy
{
    GLushort x;
    GLushort y;
    GLushort texX;
    GLushort texY;
};

/*!
 * \brief The Uploader class
 *
 * A framebuffer is bound to the
 */
class Uploader : public QObject
{
    Q_OBJECT
public:
    Uploader(QObject *parent = Q_NULLPTR);
    ~Uploader();

    QOpenGLBuffer **vertices() const;
    QOpenGLBuffer indices() const;
    GLuint **texture() const;
    unsigned int *drawCount() const;

    // VNC events
    void gotFill(const GLushort x, const GLushort y,
                 const GLushort width, const GLushort height,
                 const GLubyte red, const GLubyte green, const GLubyte blue);
    void gotCopy(const GLushort src_x, const GLushort src_y,
                 const GLushort width, const GLushort height,
                 const GLushort dest_x, const GLushort dest_y);
    void gotBitmap(const QImage &image,
                   const GLushort x, const GLushort y,
                   const GLushort width, const GLushort height);
    void gotJpeg();
    void finishedUpdate();

private:
    QOffscreenSurface m_surface;
    QOpenGLContext m_context;
    QOpenGLFunctions *m_functions;

    QOpenGLBuffer m_indices;

    QOpenGLBuffer **m_uploadVertices;
    QOpenGLBuffer **m_renderVertices;
    Vertex *m_vertexPointer;
    size_t m_vertexIndex;

    GLuint m_FBOId;
    GLuint m_texId[2];
    GLuint **m_uploadTexture;
    GLuint **m_renderTexture;

    unsigned int *m_drawCount;

    QOpenGLShaderProgram m_program;

    GLuint m_FBOTexture;

    int m_jpegs;
    // Got a FinishedUpdate event from libvnc
    bool m_finished;
    // Framebuffer swapped
    bool m_swapped;

    unsigned short m_atlasX;
    unsigned short m_atlasY;
    unsigned short m_atlasRowHeight;
    unsigned short m_atlasLastWidth;

    QOpenGLBuffer m_copyBuffer;
    Copy *m_copyPointer;
    size_t m_copyIndex;

    bool refreshAtlas(const int width, const int height);

    // Called when:
    // - got a Finished update event from libvnc
    // - and every jpeg was uploaded
    // - and the rendering thread swapped buffers
    void startRendering();
    // Called when:
    // - got a Finished update event from libvnc
    // - and the rendering thread swapped buffers
    void copyRectangles();
signals:
    // Emit from startRendering only!
    void uploadFinished();
public slots:
    void changeFBOTexture(const unsigned int FBOTexture, const QMatrix4x4 &ortho);
    void createContext(QOpenGLContext *context);
    void finishedJpeg(const unsigned char *image,
                      const GLushort x, const GLushort y,
                      const GLushort width, const GLushort height);
    void frameSwapped();
};

#endif // UPLOADER_H
